@@ -20,13 +20,12 @@ static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
 
 /* Initializes a new instance. */
 static void initialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
+    /* We're only ever called with tc->allocate_in_gen2 set, so no need to MVMROOT */
     MVMCompUnit *cu = (MVMCompUnit *)root;
-    MVMROOT(tc, cu, {
-        MVMObject *rm = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTReentrantMutex);
-        MVM_ASSIGN_REF(tc, &(root->header), cu->body.deserialize_frame_mutex, rm);
-        cu->body.inline_tweak_mutex = MVM_malloc(sizeof(uv_mutex_t));
-        uv_mutex_init(cu->body.inline_tweak_mutex);
-    });
+    MVMObject *rm = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTReentrantMutex);
+    MVM_ASSIGN_REF(tc, &(root->header), cu->body.deserialize_frame_mutex, rm);
+    cu->body.inline_tweak_mutex = MVM_malloc(sizeof(uv_mutex_t));
+    uv_mutex_init(cu->body.inline_tweak_mutex);
 }
 
 /* Copies the body of one object to another. */
@@ -64,6 +63,8 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     /* Add various other referenced strings, etc. */
     MVM_gc_worklist_add(tc, worklist, &body->hll_name);
     MVM_gc_worklist_add(tc, worklist, &body->filename);
+    MVM_gc_worklist_add(tc, worklist, &body->resolver);
+    MVM_gc_worklist_add(tc, worklist, &body->dynamic_resolver);
 }
 
 /* Called by the VM in order to free memory associated with this object. */
@@ -81,17 +82,11 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVM_free(body->inline_tweak_mutex);
     MVM_free(body->coderefs);
     if (body->callsites)
-        MVM_fixed_size_free(tc, tc->instance->fsa,
-            body->num_callsites * sizeof(MVMCallsite *),
-            body->callsites);
+        MVM_free(body->callsites);
     if (body->extops)
-        MVM_fixed_size_free(tc, tc->instance->fsa,
-            body->num_extops * sizeof(MVMExtOpRecord),
-            body->extops);
+        MVM_free(body->extops);
     if (body->strings)
-        MVM_fixed_size_free(tc, tc->instance->fsa,
-            body->num_strings * sizeof(MVMString *),
-            body->strings);
+        MVM_free(body->strings);
     MVM_free(body->scs);
     MVM_free(body->scs_to_resolve);
     MVM_free(body->sc_handle_idxs);
