@@ -79,7 +79,7 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
 
                 break;
             }
-            case MVM_OP_invoke_o:
+            case MVM_OP_dispatch_o:
             case MVM_OP_param_rp_o:
             case MVM_OP_param_rn_o:
             case MVM_OP_param_rn2_o:
@@ -126,7 +126,6 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             case MVM_OP_coerce_is:
             case MVM_OP_coerce_us:
             case MVM_OP_coerce_ns:
-            case MVM_OP_smrt_strify:
             case MVM_OP_buffertocu:
             case MVM_OP_serializetobuf:
             case MVM_OP_getsignals:
@@ -196,8 +195,6 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             case MVM_OP_clargs:
             case MVM_OP_cwd:
             case MVM_OP_newthread:
-            case MVM_OP_tryfindmeth:
-            case MVM_OP_tryfindmeth_s:
             case MVM_OP_accept_sk:
             case MVM_OP_bind_sk:
             case MVM_OP_connect_sk:
@@ -215,7 +212,6 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             case MVM_OP_gethllsym:
             case MVM_OP_getcurhllsym:
             case MVM_OP_settypecache:
-            case MVM_OP_setmethcache:
             case MVM_OP_newtype:
             case MVM_OP_nfafromstatelist:
             case MVM_OP_nfarunproto:
@@ -227,7 +223,6 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             case MVM_OP_concat_s:
             case MVM_OP_repeat_s:
             case MVM_OP_substr_s:
-            case MVM_OP_multicacheadd:
             case MVM_OP_radix:
             case MVM_OP_radix_I: {
                 add_allocation_logging(tc, g, bb, ins);
@@ -365,12 +360,14 @@ void MVM_profile_ensure_uninstrumented(MVMThreadContext *tc, MVMStaticFrame *sf)
 void MVM_profile_instrumented_start(MVMThreadContext *tc, MVMObject *config) {
     /* Wait for specialization thread to stop working, so it won't trip over
      * bytecode instrumentation, then enable profiling. */
+    MVM_gc_mark_thread_blocked(tc);
     uv_mutex_lock(&(tc->instance->mutex_spesh_sync));
     while (tc->instance->spesh_working != 0)
         uv_cond_wait(&(tc->instance->cond_spesh_sync), &(tc->instance->mutex_spesh_sync));
     tc->instance->profiling = 1;
     tc->instance->instrumentation_level++;
     uv_mutex_unlock(&(tc->instance->mutex_spesh_sync));
+    MVM_gc_mark_thread_unblocked(tc);
 }
 
 /* Simple allocation functions. */
@@ -438,6 +435,10 @@ typedef struct {
     MVMString *nursery_fresh;
     MVMString *nursery_seen;
     MVMString *gen2;
+    MVMString *mono;
+    MVMString *poly;
+    MVMString *mega;
+    MVMString *blow;
 } ProfDumpStrs;
 
 typedef struct {
@@ -572,6 +573,46 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
         /* Use static frame memory address to get a unique ID. */
         MVM_repr_bind_key_o(tc, node_hash, pds->id,
             box_i(tc, (MVMint64)(uintptr_t)sf));
+
+        /* Add data about cache entries. */
+        MVMDispInlineCache *ic = &(sf->body.inline_cache);
+        MVMuint32 monomorphic = 0;
+        MVMuint32 polymorphic = 0;
+        MVMuint32 megamorphic = 0;
+        MVMuint32 megamorphic_blowout = 0;
+        for (i = 0; i < ic->num_entries; i++) {
+            /* Skip entries that don't correspond to a bytecode operation. */
+            MVMDispInlineCacheEntry *entry = ic->entries[i];
+            MVMuint32 count = 0;
+            switch (MVM_disp_inline_cache_try_get_kind(tc, entry)) {
+                case MVM_INLINE_CACHE_KIND_MONOMORPHIC_DISPATCH:
+                case MVM_INLINE_CACHE_KIND_MONOMORPHIC_DISPATCH_FLATTENING:
+                    count = 1;
+                    break;
+                case MVM_INLINE_CACHE_KIND_POLYMORPHIC_DISPATCH:
+                    count = ((MVMDispInlineCacheEntryPolymorphicDispatch *)entry)->num_dps;
+                    break;
+                case MVM_INLINE_CACHE_KIND_POLYMORPHIC_DISPATCH_FLATTENING:
+                    count = ((MVMDispInlineCacheEntryPolymorphicDispatchFlattening *)entry)->num_dps;
+                    break;
+            }
+            if (count == 1)
+                monomorphic++;
+            else if (count > 1 && count <= 4)
+                polymorphic++;
+            else if (count > 4 && count < MVM_INLINE_CACHE_MAX_POLY)
+                megamorphic++;
+            else if (count == MVM_INLINE_CACHE_MAX_POLY)
+                megamorphic_blowout = 1;
+        }
+        if (monomorphic)
+            MVM_repr_bind_key_o(tc, node_hash, pds->mono, box_i(tc, monomorphic));
+        if (polymorphic)
+            MVM_repr_bind_key_o(tc, node_hash, pds->poly, box_i(tc, polymorphic));
+        if (megamorphic)
+            MVM_repr_bind_key_o(tc, node_hash, pds->mega, box_i(tc, megamorphic));
+        if (megamorphic_blowout)
+            MVM_repr_bind_key_o(tc, node_hash, pds->blow, box_i(tc, megamorphic_blowout));
     } else if (pcn->native_target_name) {
         MVMString *function_name_string =
             MVM_string_utf8_c8_decode(tc, tc->instance->VMString,
@@ -837,6 +878,10 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.nursery_fresh   = str(tc, "nursery_fresh");
         pds.nursery_seen    = str(tc, "nursery_seen");
         pds.gen2            = str(tc, "gen2");
+        pds.mono            = str(tc, "mono");
+        pds.poly            = str(tc, "poly");
+        pds.mega            = str(tc, "mega");
+        pds.blow            = str(tc, "blow");
 
         pds.stolen_gen2_roots  = str(tc, "stolen_gen2_roots");
         pds.has_unmanaged_data = str(tc, "has_unmanaged_data");
@@ -901,12 +946,14 @@ static MVMObject * dump_data(MVMThreadContext *tc) {
 MVMObject * MVM_profile_instrumented_end(MVMThreadContext *tc) {
 
     /* Disable profiling. */
+    MVM_gc_mark_thread_blocked(tc);
     uv_mutex_lock(&(tc->instance->mutex_spesh_sync));
     while (tc->instance->spesh_working != 0)
         uv_cond_wait(&(tc->instance->cond_spesh_sync), &(tc->instance->mutex_spesh_sync));
     tc->instance->profiling = 0;
     tc->instance->instrumentation_level++;
     uv_mutex_unlock(&(tc->instance->mutex_spesh_sync));
+    MVM_gc_mark_thread_unblocked(tc);
 
     /* Build and return result data structure. */
     return dump_data(tc);
@@ -976,12 +1023,7 @@ void MVM_profile_instrumented_free_data(MVMThreadContext *tc) {
         MVM_VECTOR_DESTROY(ptd->staticframe_array);
         MVM_VECTOR_DESTROY(ptd->type_array);
         for (MVMuint32 i = 0; i < ptd->num_gcs; i++)
-            MVM_fixed_size_free(
-                tc,
-                tc->instance->fsa,
-                ptd->gcs[i].alloc_dealloc * sizeof(MVMProfileDeallocationCount),
-                ptd->gcs[i].deallocs
-            );
+            MVM_free(ptd->gcs[i].deallocs);
         MVM_free(ptd->gcs);
         MVM_free(ptd);
         tc->prof_data = NULL;
